from flask import Flask, Request, Response

from regia.utils import SingleTon
from regia.container import Mapping


class Application(SingleTon, Flask):

    def __init__(self, *args, **kwargs):
        self.app_path: set = kwargs.pop('app_path', set())
        super(Application, self).__init__(*args, **kwargs)

    def setup(self, config):
        config.render(**self.default_config)
        from django.conf import settings
        settings.configure(config)
        self.config = config
        self._register_mapping()
        self.patch_json()
        self.auto_find_app_path()

    def _register_mapping(self):
        """注册路由模块"""
        mapping = Mapping()
        for item in mapping:
            options = item.pop('options', {})
            self.add_url_rule(**item, **options)

    @staticmethod
    def patch_json():
        try:
            import simplejson as json
        except ImportError:
            import json
        Request.json_module = json
        Response.json_module = json

    def auto_find_app_path(self):
        from regia.utils import import_string
        for path in self.app_path:
            import_string(path)


# 单例全局的application
def get_wsgi_application():
    app = Application(__name__)
    return app



from functools import wraps
from regia.container import Mapping


# 注册路由
def route(rule: str, **options):
    def wrapper(view_func):
        endpoint = options.pop("endpoint", None)
        mapping = Mapping()
        item = {'rule': rule, 'endpoint': endpoint, 'view_func': view_func, 'options': options}
        mapping.register(item)

        @wraps(view_func)
        def inner(*args, **kwargs):
            return view_func(*args, **kwargs)

        return inner

    return wrapper



from threading import Lock
from collections import UserList
from regia.utils import SingleTon


class List(UserList):
    """
    List is a collection which is thread safe.
    """

    def __init__(self, *args, **kwargs):
        super(List, self).__init__(*args, **kwargs)
        self.lock = Lock()

    def append(self, item):
        self.lock.acquire()
        super(List, self).append(item)
        self.lock.release()


class SingleTonList(SingleTon, List):
    pass


class Mapping(SingleTon):
    mapping = []
    lock = Lock()
    path = set()

    def register(self, item: dict):
        self.lock.acquire()
        self.__class__.mapping.append(item)
        self.lock.release()

    def add_path(self, package):
        self.lock.acquire()
        self.__class__.path.add(package)
        self.lock.release()

    def __iter__(self):
        for item in self.mapping:
            yield item

    def __getitem__(self, item):
        return self.mapping[item]



import json

from abc import abstractmethod

import nacos


class Config(dict):
    """Base class for regia config. All config must inherit from this class"""

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            pass

    @abstractmethod
    def initialize(self) -> dict:
        """
        配置抽象接口,所有实现类在此完成数据的读取并返回
        :return:
        """
        ...

    def render(self, **kwargs):
        data = self.initialize()
        kwargs.update(data)
        self.update(kwargs)


class JsonConfig(Config):
    """json config implement Config"""

    def __init__(self, path):
        super(JsonConfig, self).__init__()
        self.path = path

    def initialize(self) -> dict:
        with open(self.path, 'r') as f:
            return json.loads(f.read())


class NaconsConfig(Config):
    """Nacons config implement Config"""

    def __init__(self, data_id: str, group: str, **kwargs):
        super(NaconsConfig, self).__init__()
        self.server_address = kwargs.pop('server_address', 'localhost')
        self.data_id = data_id
        self.group = group
        self.kwargs = kwargs

    def initialize(self) -> dict:
        client = nacos.NacosClient(server_addresses=self.server_address, **self.kwargs)
        content = client.get_config(self.data_id, self.group)
        return json.loads(content)



from functools import wraps
from django.conf import settings
from regia.exceptions import ConfigException


def ensure_databases(func):
    """
    确保数据库配置信息正常
    prepare database
    默认的配置需要一个default的instance
    对外屏蔽这一层
    :param func: 数据库连接函数
    :return:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        databases: dict = getattr(settings, 'DATABASES', None)
        # 判断是否为空
        if not databases:
            raise ConfigException('`DATABASES` is required')
        if databases.get('default') is None:
            raise ConfigException('`DATABASES` must have a `default` key')
        return func(*args, **kwargs)

    return wrapper



from pymongo import MongoClient

from django.conf import settings

from regia.exceptions import ConfigException
from ._db import ensure_databases

clients = {}


@ensure_databases
def connection(use: str):
    try:
        return clients[use]
    except KeyError:
        databases = getattr(settings, "DATABASES")
        kwargs = databases.get(use)
        if kwargs is None:
            raise ConfigException('could not find %s database configuration from `DATABASES`' % use)
        clients[use] = MongoClient(**kwargs)
    return clients[use]



import operator
from importlib import import_module
from threading import Lock
import copy


class SingleTon(object):
    __instance = None
    __lock = Lock()

    @classmethod
    def __new__(cls, *args, **kwargs):
        """
        重写new方法，实现单例模式，便于全局调用
        :param args:
        :param kwargs:
        """
        if cls.__instance is None:
            cls.__lock.acquire()
            if cls.__instance is None:
                cls.__instance = super().__new__(cls)
            cls.__lock.release()
        return cls.__instance


empty = object()


def new_method_proxy(func):
    def inner(self, *args):
        if self._wrapped is empty:
            self._setup()
        return func(self._wrapped, *args)

    return inner


def unpickle_lazyobject(wrapped):
    return wrapped


class LazyObject:
    """
    懒加载对象
    """
    _wrapped = None

    def __init__(self):

        self._wrapped = empty

    __getattr__ = new_method_proxy(getattr)

    def __setattr__(self, name, value):
        if name == "_wrapped":
            self.__dict__["_wrapped"] = value
        else:
            if self._wrapped is empty:
                self._setup()
            setattr(self._wrapped, name, value)

    def __delattr__(self, name):
        if name == "_wrapped":
            raise TypeError("can't delete _wrapped.")
        if self._wrapped is empty:
            self._setup()
        delattr(self._wrapped, name)

    def _setup(self):
        raise NotImplementedError('subclasses of LazyObject must provide a _setup() method')

    def __reduce__(self):
        if self._wrapped is empty:
            self._setup()
        return unpickle_lazyobject, (self._wrapped,)

    def __copy__(self):
        if self._wrapped is empty:
            return type(self)()
        else:
            return copy.copy(self._wrapped)

    def __deepcopy__(self, memo):
        if self._wrapped is empty:
            result = type(self)()
            memo[id(self)] = result
            return result
        return copy.deepcopy(self._wrapped, memo)

    __bytes__ = new_method_proxy(bytes)
    __str__ = new_method_proxy(str)
    __bool__ = new_method_proxy(bool)

    __dir__ = new_method_proxy(dir)

    __class__ = property(new_method_proxy(operator.attrgetter("__class__")))
    __eq__ = new_method_proxy(operator.eq)
    __lt__ = new_method_proxy(operator.lt)
    __gt__ = new_method_proxy(operator.gt)
    __ne__ = new_method_proxy(operator.ne)
    __hash__ = new_method_proxy(hash)

    __getitem__ = new_method_proxy(operator.getitem)
    __setitem__ = new_method_proxy(operator.setitem)
    __delitem__ = new_method_proxy(operator.delitem)
    __iter__ = new_method_proxy(iter)
    __len__ = new_method_proxy(len)
    __contains__ = new_method_proxy(operator.contains)


def import_string(dotted_path):
    """
    import module by string
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError('Module "%s" does not define a "%s" attribute/class' % (
            module_path, class_name)
                          ) from err



from setuptools import setup

SET_UP_ATTRS = dict(
    name='regia',
    version='0.1.1',
    author='eat-more-apple',
    author_email='wannieqing@sailingmarvel.onaliyun.com',
    packages=['regia', 'regia.plugins'],
    url="http://106.75.251.250:8081/nieqing/regia",
    install_requires=(
        'django==3.2.7',
        'flask==2.0.1',
        'pymongo==3.12.0',
        'simplejson==3.17.5',
        'nacos-sdk-python==0.1.6',
        'mysqlclient==2.0.3',
    )
)

# 项目打包发布
setup(**SET_UP_ATTRS)
